import org.apache.regexp.*;
import java.io.*;
import java.util.*;

/** 
 * Parse a XSL-File and generate a JavaBean class containing
 * getter and setter methods for every input, checkbox and radio parameter.
 * 
 * The generated JavaBean implements java.io.Serializable
 * and - if given - an additional interface specified by
 * the "-i:<interface name>" option. The implementation for this 
 * interface follows by a file given with the "-f:<filename>" option.
 * In the text file that represents the interface implementation
 * the generated classname can be referenced using "%f%.
 * 
 * If derived from a superclass this superclass may be
 * specified using the "-e:<superclassname>" option.
 * 
 * Out of the HTML <title> tag in the XSL file a String member
 * "_title" is generated with the corresponding getter/setter method.
 * 
 * The generated filename consists of the XSL filename minus ".xsl"
 * and has the suffix "Model.java".
 *
 * Note: there is no need to try to get "better" performance using a
 * StringBuffer instead of the "+" String concatenation operator inside loops.
 * 
 * @author Klaus Berg
 */
public class XslBeanGenerator {

    // The RE Patterns
    private static final RE titlePattern = new RE("<title.+(/>|</title>)", RE.MATCH_CASEINDEPENDENT);
    private static final RE inputLinePattern = new RE("(<!--.*)|((<input.+(/>|</input>))|(<select[^>]+>)|(<textarea[^>]+>))", RE.MATCH_CASEINDEPENDENT);
    private static final RE hiddenPattern = new RE("type\\s*=\\s*\"hidden\"", RE.MATCH_CASEINDEPENDENT);
    private static final RE textPattern = new RE("type\\s*=\\s*\"text\"", RE.MATCH_CASEINDEPENDENT);                                         
    private static final RE cbPattern = new RE("type\\s*=\\s*\"checkbox\"", RE.MATCH_CASEINDEPENDENT);                                       
    private static final RE radioPattern = new RE("type\\s*=\\s*\"radio\"", RE.MATCH_CASEINDEPENDENT);                                       
    private static final RE namePattern = new RE("name\\s*=\\s*\"[^\"]+\"", RE.MATCH_CASEINDEPENDENT);                                       

    private static TreeMap parameterMap = new TreeMap();
    private static TreeMap getterSetterMethodsMap = new TreeMap();
    private static boolean generateCodeForHiddenParameters = false;


    public static void main(String[] args) throws Exception {
        if (args.length == 0) {
            System.out.println(
                              "usage: java XslBeanGenerator [-h -p:<package name> -e:<superclass> -i:<interface name> -f:<filename] <file1.xsl> <file2.xsl> ...");
            System.exit(1);
        }
        ArgumentParser parser = new ArgumentParser(args);
        generateCodeForHiddenParameters = parser.hasOption("h");
        String xslPathname = null;
        while ((xslPathname=parser.nextParam()) != null) {
            System.out.println("processing XSL file: " + xslPathname);
            parameterMap.clear();
            getterSetterMethodsMap.clear();            
            Reader r = new FileReader(xslPathname);
            // The RE package ReaderCharacterIterator, a "front end"
            // around the Reader object.
            CharacterIterator in = new ReaderCharacterIterator(r);
            int end = 0;
            String title = "";
            if (titlePattern.match(in, end)) {
                end = titlePattern.getParenEnd(0);
                String titleLine = in.substring(titlePattern.getParenStart(0), end);
                int startTitle = titleLine.indexOf('>') + 1;
                int endTitle = titleLine.lastIndexOf('<');
                title = titleLine.substring(startTitle, endTitle);
                String titleParameter = "    private String _title = \"" + title + "\";";
                parameterMap.put("_title", titleParameter);
            }
            // For each match in the input, extract and generate corresponding bean statements
            while (inputLinePattern.match(in, end)) {
                end = inputLinePattern.getParenEnd(0);
                String line = in.substring(inputLinePattern.getParenStart(0), end);
                if (line.startsWith("<!--")) {
                    continue; // ignore comments
                }
                if (line.startsWith("<input")) {
                    processInputTextLine(line);
                }
                else if (line.startsWith("<select")) {
                    processSelectLine(line);
                }
                else if (line.startsWith("<textarea")) {
                    processTextareaLine(line);
                }
            }

            File f = new File(xslPathname);
            String xslFilename = f.getName();
            String classname = firstCharToUpperCase(xslFilename.substring(0, xslFilename.length()-4) + "Model");
            String outputFilename = classname + ".java";
            System.out.println("generating output Java file: " + outputFilename);
            try {
                BufferedWriter writer = new BufferedWriter(new FileWriter(outputFilename));
                if (parser.hasOption("p")) {
                    writer.write("package " + parser.getOption("p") + ";\n\n"); 
                }
                String headerComment = 
                    "/**\n"
                    + " * JavaBean " + firstCharToUpperCase(classname) + " generated from XSL file " + xslPathname + ";\n"
                    + " * generated by XslBeanGenerator.\n"
                    + " *\n"
                    + " * @author Klaus Berg\n"
                    + " */\n";
                writer.write(headerComment);
                String additionalIF = parser.hasOption("i") ? ", " + parser.getOption("i") : "" ; 
                String additonalExtends =  parser.hasOption("e") ? " extends " + parser.getOption("e") :  "";
                writer.write("public final class " + classname + additonalExtends + " implements java.io.Serializable" + additionalIF + " {\n" + "\n");

                Set set = parameterMap.entrySet();
                Iterator iter = set.iterator();
                while (iter.hasNext()) {
                    Map.Entry me = (Map.Entry)iter.next();
                    writer.write((String)me.getValue() + "\n");
                }
                writer.write("\n\n");

                if (parser.hasOption("f")) {
                    String interfaceImplementationPath = parser.getOption("f");
                    try {
                        BufferedReader reader = new BufferedReader(new FileReader(interfaceImplementationPath));
                        String line;
                        StringBuffer sbuf = new StringBuffer(2048);
                        while ((line=reader.readLine()) != null) {
                            int substitutionIndex = line.indexOf("%f%");
                            if (substitutionIndex != -1) {
                                line = line.substring(0, substitutionIndex) + classname + line.substring(substitutionIndex+3); 
                            }
                            sbuf.append(line);
                            sbuf.append('\n');
                        }
                        getterSetterMethodsMap.put("_title", sbuf.toString());
                        reader.close();
                    }
                    catch (IOException e) {
                        e.printStackTrace();
                        System.exit(1);
                    }
                }

                set = getterSetterMethodsMap.entrySet();
                iter = set.iterator();
                while (iter.hasNext()) {
                    Map.Entry me = (Map.Entry)iter.next();
                    writer.write((String)me.getValue() + "\n");
                }
                writer.write("}" + "\n");
                writer.close();
            }
            catch (IOException e) {
                e.printStackTrace();
                System.exit(1);
            }
        }
    }

    private static void processInputTextLine(String line) {
        if (textPattern.match(line)) {
            if (namePattern.match(line)) {
                processStringParameter(line, "input-text parameter ");
            }
        }
        else if (cbPattern.match(line)) {
            if (namePattern.match(line)) {
                processBooleanParameter(line, "input-checkbox parameter ");
            }
        }
        else if (radioPattern.match(line)) {
            if (namePattern.match(line)) {
                processStringParameter(line, "input-radio parameter ");
            }
        }
        else if (generateCodeForHiddenParameters && hiddenPattern.match(line)) {
            if (namePattern.match(line)) {
                processStringParameter(line, "input-hidden parameter ");
            }
        }
    }

    private static void processSelectLine(String line) {
        if (namePattern.match(line)) {
            processStringParameter(line, "select parameter ");
        }
    }

    private static void processTextareaLine(String line) {
        if (namePattern.match(line)) {
            processStringParameter(line, "textarea parameter ");
        }
    }

    private static void processStringParameter(String line, String msg) {
        String nameLine = line.substring(namePattern.getParenStart(0), namePattern.getParenEnd(0));
        String parameterName = nameLine.substring(nameLine.indexOf('"'));
        String p = parameterName.substring(1, parameterName.length()-1);
        String javaParameterName = createJavaParameterName(p);
        String memberDeclaration;
        if (parameterName.indexOf('{') == -1) {
            memberDeclaration = "    private String " + javaParameterName + " = \"\"; // " + msg +parameterName;
        }
        else {
            String arrayParameterName = createArrayParameterName(p);
            String msg1 = msg + "ARRAY-ELEMENT ";
            memberDeclaration = "    private String " + javaParameterName + " = \"\"; // " + msg1 +parameterName;
        }
        parameterMap.put(javaParameterName, memberDeclaration);
        String getterSetterMethod = 
            "    public String get" + firstCharToUpperCase(javaParameterName) + "() {\n" 
            + "        return this." + javaParameterName + ";\n    }\n\n"
            + "    public void set" + firstCharToUpperCase(javaParameterName) + "(String newValue) {\n"
            + "        this." + javaParameterName + " = newValue;\n    }\n";
        getterSetterMethodsMap.put(javaParameterName, getterSetterMethod);
    }


    private static void processBooleanParameter(String line, String msg) {
        String nameLine = line.substring(namePattern.getParenStart(0), namePattern.getParenEnd(0));
        String parameterName = nameLine.substring(nameLine.indexOf('"'));
        String p = parameterName.substring(1, parameterName.length()-1);
        String javaParameterName = createJavaParameterName(p);
        String memberDeclaration;
        if (parameterName.indexOf('{') == -1) {
            memberDeclaration = "    private boolean " + javaParameterName + " = false; // " + msg +parameterName;
        }
        else {
            String arrayParameterName = createArrayParameterName(p);
            String msg1 = msg + "ARRAY-ELEMENT ";
            memberDeclaration = "    private boolean " + javaParameterName + " = false; // " + msg1 +parameterName;
        }
        parameterMap.put(memberDeclaration, memberDeclaration);
        String getterSetterMethod = null;
        if (javaParameterName.toLowerCase().startsWith("get")) {
            getterSetterMethod = "    public boolean " + firstCharToLowerCase(javaParameterName) + "() {\n";
        }
        else {
            getterSetterMethod = "    public boolean is" + firstCharToUpperCase(javaParameterName) + "() {\n";
        }
        getterSetterMethod += 
            "        return this." + javaParameterName + ";\n    }\n\n"
            + "    public void set" + firstCharToUpperCase(javaParameterName) + "(boolean newValue) {\n"
            + "        this." + javaParameterName + " = newValue;\n    }\n";
        getterSetterMethodsMap.put(javaParameterName, getterSetterMethod);        
    }

    private static String createJavaParameterName(String s) {
        String parameterName = s.toLowerCase();
        if (parameterName.indexOf('-') != -1) {
            parameterName = parameterName.replace('-', '_');
        }
        int i1;
        if ((i1=parameterName.indexOf('{')) != -1) {
            StringBuffer sbuf = new StringBuffer(s.length()+10);
            parameterName = parameterName.substring(0, i1-1);
        }
        return parameterName;
    }


    private static String createArrayParameterName(String s) {
        String parameterName = s.toLowerCase();        
        if (parameterName.indexOf('-') != -1) {
            parameterName = parameterName.replace('-', '_');
        }
        if (parameterName.indexOf('/') != -1) {
            parameterName = parameterName.replace('/', '_');
        }
        if (parameterName.indexOf('@') != -1) {
            parameterName = parameterName.replace('@', '_'); 
        }
        if (parameterName.indexOf('}') != -1) {
            parameterName = parameterName.replace('}', '_');
        }
        int i1;
        if ((i1=parameterName.indexOf('{')) != -1) {
            StringBuffer sbuf = new StringBuffer(s.length()+10);
            parameterName = sbuf.append("\"").append(parameterName.substring(0, i1-1)).append("\" + ").append(parameterName.substring(i1+1)).toString();
        }
        return parameterName;
    }

    private static String firstCharToUpperCase(String s) {
        return Character.toUpperCase(s.charAt(0)) + s.substring(1);
    }

    private static String firstCharToLowerCase(String s) {
        return Character.toLowerCase(s.charAt(0)) + s.substring(1);
    }
}
